
From: Maneesh Soni <maneesh@in.ibm.com>

This patch changes the way DCACHE_REFERENCED flag is used. It 
got messed up in dcache_rcu iterations. I hope this will be ok now.

The flag was meant to be advisory flag which is used while 
prune_dcache() so as not to free dentries which have recently 
entered d_lru list. At first pass in prune_dcache the dentries 
marked DCACHE_REFERENCED are left with the flag reset. and they 
are freed in the next pass.

So, now we mark the dentry as DCACHE_REFERENCED when it is first 
entering the d_lru list in dput() and resetthe flag in prune_dcache().
If the flag remains reset in the next call to prune_dcache(), the 
dentry is then freed.

Also I don't think any file system have to use this flag as it is taken
care by the dcache layer. The patch removes such code from a few of file
systems. Moreover these filesystems were anyway doing worng thing as they
were changing the flag out of dcache_lock.

Changes:
o dput() marks dentry DCACHE_REFERENCED when it is added to the dentry_unused
  list
o no need to set the flag in dget, dget_locked, d_lookup as these guys anyway
  increments the ref count.
o check the ref count in prune_dcache and use DCACHE_REFERENCED flag just for
  two stage aging.
o remove code for setting DACACHE_REFERENCED from reiserfs, fat, xfs and 
  exportfs.




 fs/dcache.c              |   20 ++++++++------------
 fs/exportfs/expfs.c      |    3 ---
 fs/fat/inode.c           |    1 -
 fs/reiserfs/inode.c      |    1 -
 fs/xfs/linux/xfs_super.c |    1 -
 include/linux/dcache.h   |    1 -
 6 files changed, 8 insertions(+), 19 deletions(-)

diff -puN fs/dcache.c~DCACHE_REFERENCED-fixes fs/dcache.c
--- 25/fs/dcache.c~DCACHE_REFERENCED-fixes	2003-04-14 12:20:24.000000000 -0700
+++ 25-akpm/fs/dcache.c	2003-04-14 12:20:24.000000000 -0700
@@ -155,12 +155,11 @@ repeat:
  	if (d_unhashed(dentry))
 		goto kill_it;
   	if (list_empty(&dentry->d_lru)) {
-  		dentry->d_vfs_flags &= ~DCACHE_REFERENCED;
+  		dentry->d_vfs_flags |= DCACHE_REFERENCED;
   		list_add(&dentry->d_lru, &dentry_unused);
   		dentry_stat.nr_unused++;
   	}
  	spin_unlock(&dentry->d_lock);
-	dentry->d_vfs_flags |= DCACHE_REFERENCED;
 	spin_unlock(&dcache_lock);
 	return;
 
@@ -250,7 +249,6 @@ int d_invalidate(struct dentry * dentry)
 static inline struct dentry * __dget_locked(struct dentry *dentry)
 {
 	atomic_inc(&dentry->d_count);
-	dentry->d_vfs_flags |= DCACHE_REFERENCED;
 	if (atomic_read(&dentry->d_count) == 1) {
 		dentry_stat.nr_unused--;
 		list_del_init(&dentry->d_lru);
@@ -379,17 +377,16 @@ static void prune_dcache(int count)
 		dentry = list_entry(tmp, struct dentry, d_lru);
 
  		spin_lock(&dentry->d_lock);
+		/* leave inuse dentries */
+ 		if (atomic_read(&dentry->d_count)) {
+ 			spin_unlock(&dentry->d_lock);
+			continue;
+		}
 		/* If the dentry was recently referenced, don't free it. */
 		if (dentry->d_vfs_flags & DCACHE_REFERENCED) {
 			dentry->d_vfs_flags &= ~DCACHE_REFERENCED;
-
-			/* don't add non zero d_count dentries 
-			 * back to d_lru list
-			 */
- 			if (!atomic_read(&dentry->d_count)) {
- 				list_add(&dentry->d_lru, &dentry_unused);
- 				dentry_stat.nr_unused++;
- 			}
+ 			list_add(&dentry->d_lru, &dentry_unused);
+ 			dentry_stat.nr_unused++;
  			spin_unlock(&dentry->d_lock);
 			continue;
 		}
@@ -1027,7 +1024,6 @@ struct dentry * __d_lookup(struct dentry
 		if (likely(move_count == dentry->d_move_count)) {
 			if (!d_unhashed(dentry)) {
 				atomic_inc(&dentry->d_count);
-				dentry->d_vfs_flags |= DCACHE_REFERENCED;
 				found = dentry;
 			}
 		}
diff -puN fs/exportfs/expfs.c~DCACHE_REFERENCED-fixes fs/exportfs/expfs.c
--- 25/fs/exportfs/expfs.c~DCACHE_REFERENCED-fixes	2003-04-14 12:20:24.000000000 -0700
+++ 25-akpm/fs/exportfs/expfs.c	2003-04-14 12:20:24.000000000 -0700
@@ -91,7 +91,6 @@ find_exported_dentry(struct super_block 
 			if (dentry != result &&
 			    acceptable(context, dentry)) {
 				dput(result);
-				dentry->d_vfs_flags |= DCACHE_REFERENCED;
 				return dentry;
 			}
 			spin_lock(&dcache_lock);
@@ -271,7 +270,6 @@ find_exported_dentry(struct super_block 
 		if (dentry != result &&
 		    acceptable(context, dentry)) {
 			dput(result);
-			dentry->d_vfs_flags |= DCACHE_REFERENCED;
 			return dentry;
 		}
 		spin_lock(&dcache_lock);
@@ -434,7 +432,6 @@ static struct dentry *export_iget(struct
 		iput(inode);
 		return ERR_PTR(-ENOMEM);
 	}
-	result->d_vfs_flags |= DCACHE_REFERENCED;
 	return result;
 }
 
diff -puN fs/fat/inode.c~DCACHE_REFERENCED-fixes fs/fat/inode.c
--- 25/fs/fat/inode.c~DCACHE_REFERENCED-fixes	2003-04-14 12:20:24.000000000 -0700
+++ 25-akpm/fs/fat/inode.c	2003-04-14 12:20:24.000000000 -0700
@@ -608,7 +608,6 @@ struct dentry *fat_get_dentry(struct sup
 		return ERR_PTR(-ENOMEM);
 	}
 	result->d_op = sb->s_root->d_op;
-	result->d_vfs_flags |= DCACHE_REFERENCED;
 	return result;
 }
 
diff -puN fs/reiserfs/inode.c~DCACHE_REFERENCED-fixes fs/reiserfs/inode.c
--- 25/fs/reiserfs/inode.c~DCACHE_REFERENCED-fixes	2003-04-14 12:20:24.000000000 -0700
+++ 25-akpm/fs/reiserfs/inode.c	2003-04-14 12:20:24.000000000 -0700
@@ -1294,7 +1294,6 @@ struct dentry *reiserfs_get_dentry(struc
 	    iput(inode);
 	    return ERR_PTR(-ENOMEM);
     }
-    result->d_vfs_flags |= DCACHE_REFERENCED;
     return result;
 }
 
diff -puN fs/xfs/linux/xfs_super.c~DCACHE_REFERENCED-fixes fs/xfs/linux/xfs_super.c
--- 25/fs/xfs/linux/xfs_super.c~DCACHE_REFERENCED-fixes	2003-04-14 12:20:24.000000000 -0700
+++ 25-akpm/fs/xfs/linux/xfs_super.c	2003-04-14 12:20:24.000000000 -0700
@@ -741,7 +741,6 @@ linvfs_get_dentry(
 		iput(inode);
 		return ERR_PTR(-ENOMEM);
 	}
-	result->d_vfs_flags |= DCACHE_REFERENCED;
 	return result;
 }
 
diff -puN include/linux/dcache.h~DCACHE_REFERENCED-fixes include/linux/dcache.h
--- 25/include/linux/dcache.h~DCACHE_REFERENCED-fixes	2003-04-14 12:20:24.000000000 -0700
+++ 25-akpm/include/linux/dcache.h	2003-04-14 12:20:24.000000000 -0700
@@ -270,7 +270,6 @@ static inline struct dentry *dget(struct
 		if (!atomic_read(&dentry->d_count))
 			BUG();
 		atomic_inc(&dentry->d_count);
-		dentry->d_vfs_flags |= DCACHE_REFERENCED;
 	}
 	return dentry;
 }

_
